//
// Copyright 2017 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
//

#import <SignalServiceKit/BaseModel.h>

NS_ASSUME_NONNULL_BEGIN

@class Contact;
@class SignalRecipient;
@class SignalServiceAddress;

/// A ``SignalAccount`` represents a "system contact" that is also a registered
/// Signal user. The system contact may be from the local device's address book,
/// or synced to a linked device from the primary device's address book.
///
/// System contacts with multiple Signal accounts (e.g., multiple registered
/// phone numbers) will correspond to multiple instances of ``SignalAccount``.
///
/// Note that this description applies to _persisted_ ``SignalAccount``s, which
/// will always be backed by a non-`nil` ``Contact``. In-memory
/// ``SignalAccount``s may be created for other purposes, and will not contain
/// a ``Contact``.
@interface SignalAccount : BaseModel

/// An E164 value identifying the Signal account.
@property (nullable, nonatomic, readonly) NSString *recipientPhoneNumber;

/// A UUID identifying the Signal account.
@property (nullable, nonatomic, readonly) NSString *recipientUUID;

/// An address representing the Signal account. This will be
/// the UUID, if defined, otherwise it will be the E164 number.
@property (nonatomic, readonly) SignalServiceAddress *recipientAddress;

/// This property is non-`nil` for persisted ``SignalAccount``s. It may be
/// `nil` for in-memory accounts.
@property (nonatomic, nullable, readonly) Contact *contact;

/// ``contactAvatarHash`` is the hash of the original avatar
/// data (if any) from the system contact.  We use it for
/// change detection.
///
/// This property is optional and will not be set for
/// non-contact account.
@property (nonatomic, nullable, readonly) NSData *contactAvatarHash;

/// This property is obsolete.
@property (nonatomic, nullable, readonly) NSData *contactAvatarJpegDataObsolete;

/// For contacts with more than one signal account,
/// this is a label for the account.
@property (nonatomic, readonly) NSString *multipleAccountLabelText;

- (nullable NSString *)contactPreferredDisplayName;
- (nullable NSString *)contactFullName;
- (nullable NSString *)contactFirstName;
- (nullable NSString *)contactLastName;
- (nullable NSString *)contactNicknameIfAvailable;
- (nullable NSPersonNameComponents *)contactPersonNameComponents;

+ (instancetype)new NS_UNAVAILABLE;
- (instancetype)init NS_UNAVAILABLE;
- (nullable instancetype)initWithCoder:(NSCoder *)coder NS_DESIGNATED_INITIALIZER;
- (instancetype)initWithUniqueId:(NSString *)uniqueId NS_UNAVAILABLE;
- (instancetype)initWithGrdbId:(int64_t)grdbId uniqueId:(NSString *)uniqueId NS_UNAVAILABLE;

// Convenience initializer which is neither "designated" nor "unavailable".
- (instancetype)initWithSignalRecipient:(SignalRecipient *)signalRecipient
                                contact:(nullable Contact *)contact
                      contactAvatarHash:(nullable NSData *)contactAvatarHash
               multipleAccountLabelText:(nullable NSString *)multipleAccountLabelText;

// Convenience initializer which is neither "designated" nor "unavailable".
- (instancetype)initWithSignalServiceAddress:(SignalServiceAddress *)address NS_SWIFT_NAME(init(address:));

// Convenience initializer which is neither "designated" nor "unavailable".
- (instancetype)initWithSignalServiceAddress:(SignalServiceAddress *)serviceAddress
                                     contact:(nullable Contact *)contact
                    multipleAccountLabelText:(nullable NSString *)multipleAccountLabelText;

- (instancetype)initWithSignalServiceAddress:(SignalServiceAddress *)serviceAddress
                                     contact:(nullable Contact *)contact
                           contactAvatarHash:(nullable NSData *)contactAvatarHash
                    multipleAccountLabelText:(nullable NSString *)multipleAccountLabelText NS_DESIGNATED_INITIALIZER;

- (instancetype)initWithContact:(nullable Contact *)contact
              contactAvatarHash:(nullable NSData *)contactAvatarHash
       multipleAccountLabelText:(NSString *)multipleAccountLabelText
           recipientPhoneNumber:(nullable NSString *)recipientPhoneNumber
                  recipientUUID:(nullable NSString *)recipientUUID NS_DESIGNATED_INITIALIZER;

// --- CODE GENERATION MARKER

// This snippet is generated by /Scripts/sds_codegen/sds_generate.py. Do not manually edit it, instead run `sds_codegen.sh`.

// clang-format off

- (instancetype)initWithGrdbId:(int64_t)grdbId
                      uniqueId:(NSString *)uniqueId
                         contact:(nullable Contact *)contact
               contactAvatarHash:(nullable NSData *)contactAvatarHash
   contactAvatarJpegDataObsolete:(nullable NSData *)contactAvatarJpegDataObsolete
        multipleAccountLabelText:(NSString *)multipleAccountLabelText
            recipientPhoneNumber:(nullable NSString *)recipientPhoneNumber
                   recipientUUID:(nullable NSString *)recipientUUID
NS_DESIGNATED_INITIALIZER NS_SWIFT_NAME(init(grdbId:uniqueId:contact:contactAvatarHash:contactAvatarJpegDataObsolete:multipleAccountLabelText:recipientPhoneNumber:recipientUUID:));

// clang-format on

// --- CODE GENERATION MARKER

- (BOOL)hasSameContent:(SignalAccount *)other;

- (void)updateWithContact:(nullable Contact *)contact
              transaction:(SDSAnyWriteTransaction *)transaction NS_SWIFT_NAME(updateWithContact(_:transaction:));

#if TESTABLE_BUILD
- (void)replaceContactForTests:(nullable Contact *)contact NS_SWIFT_NAME(replaceContactForTests(_:));
#endif

@end

NS_ASSUME_NONNULL_END
